;; Problem 165
;; 27 October 2007

;; A segment is uniquely defined by its two endpoints.
;; By considering two line segments in plane geometry there are three possibilities:
;; the segments have zero points, one point, or infinitely many points in common.

;; Moreover when two segments have exactly one point in common it might be the case that that common point is an endpoint of either one of the segments or of both. If a common point of two segments is not an endpoint of either of the segments it is an interior point of both segments.
;; We will call a common point T of two segments L1 and L2 a true intersection point of L1 and L2 if T is the only common point of L1 and L2 and T is an interior point of both segments.

;; Consider the three segments L1, L2, and L3:

;; L1: (27, 44) to (12, 32)
;; L2: (46, 53) to (17, 62)
;; L3: (46, 70) to (22, 40)

;; It can be verified that line segments L2 and L3 have a true intersection point. We note that as the one of the end points of L3: (22,40) lies on L1 this is not considered to be a true point of intersection. L1 and L2 have no common point. So among the three line segments, we find one true intersection point.

;; Now let us do the same for 5000 line segments. To this end, we generate 20000 numbers using the so-called "Blum Blum Shub" pseudo-random number generator.

;; s0 = 290797

;; sn+1 = sn X sn (modulo 50515093)

;; tn = sn (modulo 500)

;; To create each line segment, we use four consecutive numbers tn. That is, the first line segment is given by:

;; (t1, t2) to (t3, t4)

;; The first four numbers computed according to the above generator should be: 27, 144, 12 and 232. The first segment would thus be (27,144) to (12,232).

;; How many distinct true intersection points are found among the 5000 line segments?

;; (declaim (optimize speed))

(defvar *blum-blum-shub-start* 290797)
(defvar *blum-blum-shub-modulo* 50515093)

(defun make-blum-blum-shub-generator ()
  (let ((current *blum-blum-shub-start*))
    (lambda ()
      (let ((next (mod (* current current) *blum-blum-shub-modulo*)))
	(setf current next)
	(mod next 500)))))

(let ((gen (make-blum-blum-shub-generator)))
  (loop repeat 5
	do
	(format t "~a~&" (funcall gen))))

(defun make-point-generator ()
  (let ((gen (make-blum-blum-shub-generator)))
    (lambda ()
      (list (funcall gen)
	    (funcall gen)))))


'(let ((gen (make-point-generator)))
  (loop repeat 5
	do
	(format t "~a~&" (funcall gen))))

(declaim (inline pt.x pt.y lseg.start lseg.end))
(defun pt.x (pt) (declare (optimize speed) (type fixnum)) (first pt))
(defun pt.y (pt) (declare (optimize speed) (type fixnum)) (second pt))
(defun lseg.start (seg) (declare (optimize speed)) (first seg))
(defun lseg.end   (seg) (declare (optimize speed)) (second seg))

(defun make-line-seg-gen ()
  (let ((gen (make-point-generator)))
    (lambda ()
      (sort 
       (list (funcall gen)
	     (funcall gen))
       #'(lambda (left right)
	   (cond ((< (pt.x left) (pt.x right))
		  t)
		 ((= (pt.x left) (pt.x right))
		  (< (pt.y left) (pt.y right)))
		 (t
		  nil)))))))

'(let ((gen (make-line-seg-gen)))
  (loop repeat 5
	do
	(format t "~a~&" (funcall gen))))


(defvar *num-segments* 5000)

(defvar *line-segments*
  (let ((gen (make-line-seg-gen)))
    (loop repeat *num-segments*
	  collect (funcall gen))))

;; (length *line-segments*)
;; (first *line-segments*)

(require 'vecto)

(defun plot-lines (file-name lines &optional (red-lines (list)) (blue-lines (list)))
  (let ((width 1000)
	(height 1000))
    (vecto:with-canvas (:width width :height height)

      (vecto:set-rgb-stroke 0 0 0)
      (vecto:set-line-width 2)

      (let ((scale 2))
	(loop for line-seg in lines 
	      do
	      (vecto:move-to (* scale (pt.x (lseg.start line-seg))) (* scale (pt.y (lseg.start line-seg))))
	      (vecto:line-to (* scale (pt.x (lseg.end line-seg)))   (* scale (pt.y (lseg.end line-seg)))))
	(vecto:stroke)
	(vecto:set-rgb-stroke 0 0 1)
	(loop for line-seg in blue-lines
	      do
	      (vecto:move-to (* scale (pt.x (lseg.start line-seg))) (* scale (pt.y (lseg.start line-seg))))
	      (vecto:line-to (* scale (pt.x (lseg.end line-seg)))   (* scale (pt.y (lseg.end line-seg)))))
	(vecto:stroke)
	(vecto:set-rgb-stroke 1 0 0)
	(loop for line-seg in red-lines
	      do
	      (vecto:move-to (* scale (pt.x (lseg.start line-seg))) (* scale (pt.y (lseg.start line-seg))))
	      (vecto:line-to (* scale (pt.x (lseg.end line-seg)))   (* scale (pt.y (lseg.end line-seg)))))
	(vecto:stroke)
	(vecto:save-png file-name)))))


;; (plot-lines "/home/mortis/personal/projects/project-euler/lines.png" (subseq *line-segments* 0 500))
;; (plot-lines "/home/mortis/personal/projects/project-euler/lines-all.png" *line-segments*)

;; find the # of intersections where the intersect is not an end point
;; in common between the two segments...

(defun segments-might-overlap (seg cand)
  (or
   (and
    (or
     (< (pt.x (lseg.start seg))
	(pt.x (lseg.start cand))
	(pt.x (lseg.end   seg)))
     (< (pt.x (lseg.start seg))
	(pt.x (lseg.end   cand))
	(pt.x (lseg.end   seg))))
    (or
     (< (pt.y (lseg.start seg))
	(pt.y (lseg.start cand))
	(pt.y (lseg.end   seg)))
     (< (pt.y (lseg.start seg))
	(pt.y (lseg.end   cand))
	(pt.y (lseg.end   seg)))))
   (and
    (or
     (< (pt.x (lseg.start cand))
	(pt.x (lseg.start seg))
	(pt.x (lseg.end   cand)))
     (< (pt.x (lseg.start cand))
	(pt.x (lseg.end   seg))
	(pt.x (lseg.end   cand))))
    (or
     (< (pt.y (lseg.start cand))
	(pt.y (lseg.start seg))
	(pt.y (lseg.end   cand)))
     (< (pt.y (lseg.start cand))
	(pt.y (lseg.end   seg))
	(pt.y (lseg.end   cand)))))))


(defun find-segment-candidates (seg)
  (let ((res (list)))
    (loop for cand in *line-segments*
	  do
	  (when (segments-might-overlap seg cand)
	    (push cand res)))
    res))

(defun find-line-coeficients (pt1 pt2)
  (let* ((a (- (second pt2) (second pt1)))
	 (b (- (first pt1) (first pt2)))
	 (c (+ (* a (first pt1))
	       (* b (second pt1)))))
    (list a b c)))

;; NB: for optimization we can memoize this into *line-segments* as
;; the third element...
(defun find-line-intersect-point (seg1 seg2)
  (let ((l1 (apply #'find-line-coeficients seg1))
	(l2 (apply #'find-line-coeficients seg2)))
    (let ((det (- (* (first l1) (second l2))
		  (* (first l2) (second l1)))))
      (cond ((= 0 det)
	     nil)
	    (t
	     (list
	      (/
	       (- (* (second l2) (third l1))
		  (* (second l1) (third l2)))
	       det)
	      (/
	       (- (* (first l1) (third l2))
		  (* (first l2) (third l1)))
	       det)))))))

(defun point-is-within (pt start end)
  (and
   (or
    (< (pt.x start) (pt.x pt) (pt.x end))
    (> (pt.x start) (pt.x pt) (pt.x end)))
   (or
    (< (pt.y start) (pt.y pt) (pt.y end))
    (> (pt.y start) (pt.y pt) (pt.y end)))))

;; (point-is-within '(-1 -1) '(0 0) '(3 3))

(defun find-line-intersect (seg1 seg2)
  (let ((isect (find-line-intersect-point seg1 seg2)))
    (cond ((null isect)
	   nil)
	  ((and (point-is-within isect (lseg.start seg1) (lseg.end seg1))
		(point-is-within isect (lseg.start seg2) (lseg.end seg2)))
	   ;; the isect has to be between the start and end of both
	   ;; segments...
;; 	   (format t "x coords seg1:~a < isect:~a < seg2:~a~&"
;; 		   seg1
;; 		   isect
;; 		   seg2)
	   isect))))

;; should be none
(find-line-intersect '((0 0) (0 5))
		     '((0 1) (0 4)))

;; should be at (2.5 2.5)
(find-line-intersect '((0 0) (5 5))
		     '((5 0) (0 5)))

;; should be none (lines to intersect, but not within the segments!)
(find-line-intersect-point '((0 0) (5 5)) '((5 0) (4 1)))
(find-line-intersect '((0 0) (5 5)) '((5 0) (4 1)))

(find-line-intersect '((27 44) (12 32)) '((46 53) (17 62)))
(find-line-intersect '((27 44) (12 32)) '((46 70) (22 40)))
(find-line-intersect '((46 53) (17 62)) '((46 70) (22 40)))


(defmacro awhen (test &body body)
  `(let ((it ,test))
    (when it ,@body)))

(defvar *points-found* (make-hash-table :test #'equal))

;; (length *line-segments*)
(let ((tot-overlaps 0))
  (setf *points-found* (make-hash-table :test #'equal))
  (loop for seg in
	;; (subseq *line-segments* 0 4)
        *line-segments*
	do
	(let ((candidates (find-segment-candidates seg)))
;;	  (incf tot-overlaps (length candidates))
;; 	  (format t "possible overlaps(~a): ~a~&"
;; 		  (length candidates)
;; 		  		  seg)
	  (loop for candidate in candidates
		do
		(awhen (find-line-intersect seg candidate)
		  (incf tot-overlaps)
		  (setf (gethash it *points-found*) t)))))
		  ;; (format t "found isect:~a~&" it)))))
  (format t "tot-overlaps:~a avg:~a~&"
	  tot-overlaps
	  (/ tot-overlaps 5000.0))
  (princ *points-found*)
  (format t "*points-found* length => ~a~&"
	  (hash-table-count *points-found*)))

;; tot-overlaps:2525044 avg:505.0088
;; #<HASH-TABLE :TEST EQUAL :COUNT 1262493 {AA74C01}>*points-found* length => 1262493

(defvar *intersection-points* nil)

(progn
  (setf *intersection-points*
	(loop for key being the hash-keys of *points-found*
	      collect key))
  t)

; (subseq  *intersection-points* 0 10)


;; (hash-table-count *points-found*)
(loop for seg in *line-segments*
      do
      (remhash (lseg.start seg) *points-found*)
      (remhash (lseg.end seg) *points-found*))

;; (hash-table-count *points-found*);; 1262487

; (subseq  *intersection-points* 0 10)
;; (nth 3 *line-segments*)

(let ((idx 5))
  (loop for seg in
	(list (nth idx *line-segments*))
	do
	(let* ((candidates   (find-segment-candidates seg))
	       (overlaps     (remove-if #'(lambda (cand) (not (find-line-intersect seg cand))) candidates))
	       (non-overlaps (remove-if #'(lambda (cand)      (find-line-intersect seg cand))  candidates)))
	  (plot-lines "/home/mortis/personal/projects/project-euler/test.png" non-overlaps (list seg) overlaps))))


'(loop for seg in
      '(((0 0) (100 100))
	 ((0 100) (100 0)))
      do
      (let* ((candidates   (find-segment-candidates seg))
	     (overlaps     (remove-if #'(lambda (cand) (not (find-line-intersect seg cand))) candidates))
	     (non-overlaps (remove-if #'(lambda (cand)      (find-line-intersect seg cand))  candidates)))
	(plot-lines "/home/mortis/personal/projects/project-euler/test.png" (cons '((0 0) (500 500)) non-overlaps) (list seg) overlaps)))
